using System;
using System.IO;
using System.Xml.Serialization;
using RedRat.AVDeviceMngmt;
using RedRat.IR;

namespace RedRat.Samples.CustomSignal
{
    /// <summary>
    /// This sample shows how to create a custom IR signal.
    /// </summary>
    class CustomSignal
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main( string[] args )
        {
            new CustomSignal();
        }

        public CustomSignal()
        {
            // == Create a couple of IR signals ==
            var sigData = new byte[] {0, 1, 2, 1, 2, 2, 2, 2, 2, 1, 2};
            var sig1 = CreateSignal( "Play", sigData, sigData );
            Console.WriteLine( sig1 );

            sigData = new byte[] {0, 1, 2, 1, 2, 1, 2, 2, 2, 1, 2};
            var sig2 = CreateSignal( "Pause", sigData, sigData );
            Console.WriteLine( sig1 );

            // == Create an AV Device object so that we can then store the signal data in an XML file. ==
            var avDeviceDB = new AVDeviceDB();

            var avDevice = new AVDevice( "Sample Device", AVDevice.AVDeviceType.SET_TOP_BOX );
            avDeviceDB.AddAVDevice( avDevice );

            // Add the signals to the "Sample Device".
            avDevice.AddSignal( sig1, false );
            avDevice.AddSignal( sig2, false );

            // Store this as an XML file...
            var ser = new XmlSerializer( typeof( AVDeviceDB ) );
            TextWriter writer = new StreamWriter( ( new FileInfo( "Sample-AVDeviceDB.xml") ).FullName );
            ser.Serialize( writer, avDeviceDB );
            writer.Close();

            Console.WriteLine( "Saved IR data to file..." );
        }

        protected ModulatedSignal CreateSignal( string name, byte[] mainSignal, byte[] repeatSignal )
        {
            // Create a signal object - set its name and modulation/carrier frequency.
            var modSig = new ModulatedSignal
                             {
                                 Name = name, 
                                 ModulationFreq = 38000
                             };

            // Create lengths - ms.
            var lengths = new double[3];
            lengths[0] = 4.0;
            lengths[1] = 1.2;
            lengths[2] = 0.6;
            modSig.Lengths = lengths;

            // Pause time between main and repeats signals
            modSig.IntraSigPause = 40; // 40ms

            // No of repeats of the repeat section
            modSig.NoRepeats = 1;

            // The signal data
            if ( ( mainSignal == null ) || ( repeatSignal == null ) )
            {
                throw new Exception( "Main or repeat signals cannot be null" );
            }
            var sigData = new byte[mainSignal.Length + repeatSignal.Length + 2];
            var sigDataCount = 0;
            // Add main signal data
            for ( var i = 0; i < mainSignal.Length; i++, sigDataCount++ )
            {
                sigData[sigDataCount] = mainSignal[i];
            }
            // Mark end of main signal
            sigData[sigDataCount++] = ModulatedSignal.EOS_MARKER;
            // Add repeat signal data
            for ( var i = 0; i < mainSignal.Length; i++, sigDataCount++ )
            {
                sigData[sigDataCount] = repeatSignal[i];
            }
            // Mark end of repeat signal
            sigData[sigDataCount] = ModulatedSignal.EOS_MARKER;

            modSig.SigData = sigData;

            return modSig;
        }
    }
}
