﻿using System;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Xml.Serialization;
using RedRat.AVDeviceMngmt;
using RedRat.RedRat3;
using RedRat.RedRat4.StreamIO.X;
using RedRat.Util;

namespace RedRat.Samples.KeyPressLogger
{
    /// <summary>
    /// This application will capture incoming IR signals from a USB RedRat-X, attempt to decode them
    /// and the print out the results.
    /// 
    /// To send the log output to a file, use this:
    ///        KeyPressLogger.exe "RedRat-X 1" DeviceDB.xml > test.log
    /// 
    /// Using Powershell, output can be sent to the console AND log file like this:
    ///        KeyPressLogger.exe "RedRat-X 1" DeviceDB.xml | Tee-Object -file test.log
    /// </summary>
    public class Program
    {
        private IRedRat3 redRatX;
        private AVDeviceDB avDeviceDB;
        private bool oneshot = false;
        private readonly ManualResetEvent completionWaitHandle = new ManualResetEvent( false );

        static void Main( string[] args )
        {
            new Program( args );
        }

        public Program( string[] args )
        {
            try
            {
                CheckArgs( args );
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( $"\nError from Key Press Logger: {ex.Message}\n" );
            }
            finally
            {
                TidyUp();
            }
        }

        private void CheckArgs( string[] args )
        {
            // Need two - RR3 to use and device DB file.
            if ( args.Length < 2 )
            {
                PrintUsage();
                return;
            }

            if ( !FindRedRatX( args[0] ) ) return;
            if ( !LoadSignalData( args[1] ) ) return;

            if ( args.Length > 2 )
            {
                if ( args[2].Equals( "-oneshot", StringComparison.CurrentCultureIgnoreCase ) )
                {
                    oneshot = true;
                }
            }

            avDeviceDB.DecoderLogEnabled = false; // false = default. Set to true to turn on some decoder logging.
            StartLogging( args[0], args[1] );
        }

        /// <summary>
        /// Handles the user pressing CTRL-C to exit the application. If the RC detector is enabled
        /// when the application stops it can cause the RR3 to hang.
        /// </summary>
        private void ConsoleCancelKeyPress( object sender, ConsoleCancelEventArgs e )
        {
            TidyUp();
        }

        private bool FindRedRatX( string rrxName )
        {
            // Lookup using the 'friendly' name, e.g. "RedRatX-0".
            var usbRrxs = RedRatXUsbDevice.FindDevices().OfType<RedRatXUsbLocationInfo>().ToList();
            ;
            var li = usbRrxs.FirstOrDefault( r => r.Name.Equals( rrxName, StringComparison.CurrentCultureIgnoreCase ) );
            if ( li != null )
            {
                redRatX = RedRatX.GetInstance( li );
                return true;
            }

            // OK, haven't found an RRX so try using the device name.
            li = usbRrxs.FirstOrDefault( r => r.DeviceName.Equals( rrxName, StringComparison.CurrentCultureIgnoreCase ) );
            if ( li != null )
            {
                redRatX = RedRatX.GetInstance( li );
                return true;
            }

            PrintUsage();
            WriteError( $"Failed to find to RedRat-X '{rrxName}'" );
            WriteError( PrintFoundRedRaXs() );
            return false;
        }

        private bool LoadSignalData( string fileName )
        {
            var ser = new XmlSerializer( typeof( AVDeviceDB ) );
            FileStream fs = null;
            try
            {
                fs = new FileStream( ( new FileInfo( fileName ) ).FullName, FileMode.Open );
                avDeviceDB = (AVDeviceDB)ser.Deserialize( fs );
            }
            catch ( Exception ex )
            {
                WriteError( $"Error loading signal database file '{fileName}'\n   {ex.Message}" );
                return false;
            }
            finally
            {
                fs?.Close();
            }
            return true;
        }

        private void LogEvent( string eventMsg )
        {
            // Prints out to 100ths of a second.
            Console.WriteLine( "[{0:dd/MM/yyyy HH:mm:ss.ff}] {1}", DateTime.Now, eventMsg );
        }

        private void PrintUsage()
        {
            var version = Assembly.GetExecutingAssembly().GetName().Version;
            Console.WriteLine( "\nRedRat Key Press Logger - Version {0}.{1}", version.Major, version.Minor );
            Console.WriteLine( "-------------------------------------" );
            Console.WriteLine( "\nUsage: KeyPressLogger RedRatX-Name IR-Signal-DB-File [-oneshot]" );
            Console.WriteLine( "\n  Where:\n" );
            Console.WriteLine( "    RedRatX-Name      - Either the friendly name or USB device name." );
            Console.WriteLine( "    IR-Signal-DB-File - File containing IR signal XML data for decoding." );
            Console.WriteLine( "    -oneshot          - [Optional] Program will terminate after IR signal input.\n");
        }

        /// <summary>
        /// Handles incoming IR signal events from RR3.
        /// </summary>
        private void RedRatXOnRcDetectorSignalIn( object sender, EventArgs eventArgs )
        {
            try
            {
                var siea = eventArgs as SignalEventArgs;
                if ( siea == null ) return;

                switch ( siea.Action )
                {
                    case SignalEventAction.RC_DETECTOR_ENABLED:
                        LogEvent( "Signal input enabled" );
                        break;

                    case SignalEventAction.RC_DETECTOR_DISABLED:
                        LogEvent( "Signal input terminated" );
                        break;

                    case SignalEventAction.MODULATED_SIGNAL:
                        var sigKey = avDeviceDB.DecodeSignal( siea.ModulatedSignal );
                        var msg = sigKey.KeyStatus == SignalKey.Status.KNOWN
                            ? $"KEY: {sigKey.AVDevice.Name} -> {sigKey.Signal.Name}"
                            : "KEY: Unknown";
                        LogEvent( msg );
                        completionWaitHandle.Set(); // For use if "oneshot"
                        break;

                    case SignalEventAction.EXCEPTION:
                        LogEvent( "Error: " + siea.Exception.Message );
                        break;
                }
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( "Error decoding IR signal:\n" + ex );
            }
        }

        private void StartLogging( string rrxName, string dbFilename )
        {
            // Setup CTRL-C handler.
            Console.CancelKeyPress += ConsoleCancelKeyPress;

            // Connect to RedRat-X
            redRatX.Connect();

            // Enable IR input...
            redRatX.RCDetectorSignalIn += RedRatXOnRcDetectorSignalIn;

            redRatX.RCDetectorEnabled = true;

            if ( oneshot )
            {
                Console.WriteLine("\nRedRat Key Logger started. Waiting for one IR signal only.");

                // Wait for one signal only
                completionWaitHandle.WaitOne();
            }
            else
            {
                Console.WriteLine("\nRedRat Key Logger Started...\n");
                Console.WriteLine("   RedRat-X          : {0}", rrxName);
                Console.WriteLine("   Signal DB Filename: {0}\n", dbFilename);
                Console.WriteLine( " Hit <RETURN> to terminate...\n" );
                Console.ReadLine();
            }
        }

        private void TidyUp()
        {
            if ( redRatX == null ) return;
            if ( redRatX.RCDetectorEnabled )
            {
                redRatX.RCDetectorEnabled = false;
            }
            redRatX.Disconnect();
            RRUtil.DisposeOfAllRedRats();
        }

        private void WriteError( string message )
        {
            var currentColor = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Red;
            Console.Error.WriteLine( message );
            Console.ForegroundColor = currentColor;
        }

        private string PrintFoundRedRaXs()
        {
            var lis = RedRatXUsbDevice.FindDevices().OfType<RedRatXUsbLocationInfo>().ToList();
            if ( lis.Count == 0 )
            {
                return "No USB RedRat-X devices found.";
            }
            var sb = new StringBuilder();
            foreach ( var li in lis )
            {
                sb.Append( $"\t- Name='{li.Name}', Device name='{li.DeviceName}'" );
            }

            return sb.ToString();
        }
    }
}
