﻿using System;
using System.IO;
using System.Linq;
using System.Xml.Serialization;
using RedRat.RedRat3;
using RedRat.Util;

namespace RedRat.Samples.MacroSample
{
    /// <summary>
    /// This sample demonstrates the use of Macros in code; reading a macro file, selecting one
    /// and then outputting it via a RedRat3-II device.
    /// 
    /// The macro file has been created using the "Export" button in the macro editor.
    /// </summary>
    public class MacroSample
    {
        static void Main( string[] args )
        {
            try
            {
                new MacroSample();
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( "Problem in application: " + ex.Message );
            }
            finally
            {
                Console.WriteLine( "Hit <RETURN> to exit..." );
                Console.ReadLine();
            }
        }

        public MacroSample()
        {
            Console.WriteLine( "RedRat Macro Sample" );
            Console.WriteLine( "-------------------\n" );

            // == 1. Does the macro file exist? ==
            const string macroFileName = "MacroList.xml";
            var fileInfo = new FileInfo( macroFileName );
            if ( !fileInfo.Exists )
            {
                Console.WriteLine( $"Macro file '{macroFileName}' not found" );
                return;
            }

            // == 2. Read the macro file ==
            var macros = ReadMacroFile( macroFileName );

            // List the macros
            Console.WriteLine( $"Contains {macros.Length} macros:" );
            foreach ( var macro in macros )
            {
                Console.WriteLine( " >> {0}", macro.Name );
            }

            Console.WriteLine( "" );
            // If no macros in list, then quit
            if ( !macros.Any() ) return;

            // == 3. Output via a RedRat3-II device. This assumes one device is plugged into the PC. ==
            using ( var rr3 = FindRedRat3() )
            {
                var macro = macros[0]; // Just take the first
                macro.Execute( rr3 );
                Console.WriteLine( $"Macro '{macro.Name}' output via RedRat '{rr3.LocationInformation.Name}'.\n" );
            }
        }

        /// <summary>
        /// Reads in a macro file which has been created using the "Export" function in the Macro Editor.
        /// </summary>
        private Macro.Macro[] ReadMacroFile( string macroFileName )
        {
            FileStream fs = null;
            try
            {
                var ser = new XmlSerializer( typeof( Macro.Macro[] ) );
                fs = new FileStream( ( new FileInfo( macroFileName ) ).FullName, FileMode.Open );
                var macroList = (Macro.Macro[]) ser.Deserialize( fs );
                Console.WriteLine( $"Macro file '{macroFileName}' loaded.\n" );
                return macroList;
            }
            finally
            {
                fs?.Close();
            }
        }

        /// <summary>
        /// Simply finds the first RedRat3 attached to this computer.
        /// </summary>
        private IRedRat3 FindRedRat3()
        {
            var rr3li = RRUtil.FindRedRats( LocationInfo.RedRatType.RedRat3 ).FirstOrDefault();

            if ( rr3li == null )
            {
                throw new Exception( "Unable to find any RedRat3 devices on this computer." );
            }

            return rr3li.GetRedRat() as IRedRat3;
            ;
        }
    }
}
