using System;
using System.Text;
using System.Threading;
using System.Reflection;
using System.Xml.Serialization;

using RedRat.IR;
using RedRat.Util;
using RedRat.RedRat3;
using RedRat.RedRat3.USB;
using RedRat.AVDeviceMngmt;

//
// A basic command line interface to the RedRat3. 
//
namespace RedRat.Samples.RR_CLI
{

    /// <summary>
    /// Command line interface to RedRat operations. Used for scripting,
    /// web servers etc.
    /// </summary>
    class RR_CLI : ISerializerHandler
    {

        /// <summary>
        /// The RedRat we use for operations.
        /// </summary>
        protected IRedRat3 redRat = null;

        private readonly ArgumentParser argParser = new ArgumentParser();

        protected bool haveSignal = false;  // Set true once signal (or ex) input

        protected IRPacket irPacket = null;

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main( string[] args )
        {
            // If calling from a web server etc. - need to ensure that it does not
            // need access to the desktop.
            RedRat3USBImpl.IS_SERVER_PROCESS = true;
            new RR_CLI( args );
        }

        public RR_CLI( string[] args )
        {
            try
            {
                if ( argParser.ParseArgs( args, out var message ) )
                {
                    if ( GetRedRat() )
                    {
                        Run();
                    }
                    else
                    {
                        Console.WriteLine( $"Failed to connect to RedRat {argParser.RedRatName}" );
                        Console.WriteLine( RRUtil.PrintFoundRedRats() );
                    }
                }
                else
                {
                    PrintUsage( message );
                }
            }
            catch ( Exception ex )
            {
                Console.WriteLine( "\nERROR: " + ex.Message );
                Console.WriteLine( ex.StackTrace + "\n" );
            }
        }

        bool GetRedRat()
        {
            try
            {
                // Lookup using the device name, e.g. "RedRat-0".
                redRat = RedRat3USBImpl.GetInstance( argParser.RedRatName );
            }
            catch
            {
                // OK, haven't found an RR3 by using the device name, so try using the location name.
                var li = RRUtil.RedRatNameToLocInfo( argParser.RedRatName );
                // Only support RR3.
                if ( li != null )
                {
                    redRat = (IRedRat3)li.GetRedRat();
                }
            }
            return redRat != null;
        }

        protected void Run()
        {
            if ( argParser.ShouldBlink )
            {
                redRat.Blink();
                Console.WriteLine( "BLINK" );
            }
            if ( argParser.ShouldCapture )
            {
                CaptureSignal();
            }
            if ( argParser.ShouldOutput )
            {
                OutputSignal();
            }
        }

        bool IsXmlFile( string fileName )
        {
            var extension = fileName.Substring( startIndex: fileName.Length - 3 );
            return string.Equals( extension, "xml", StringComparison.InvariantCultureIgnoreCase );
        }

        protected void CaptureSignal()
        {
            // Capture a signal - take the next arg as filename.
            if ( !( redRat is RedRat3USBImpl ) ) return;

            var redRat3 = (RedRat3USBImpl)redRat;
            redRat3.LearningSignalIn += SignalDataHandler;
            redRat3.GetModulatedSignal( 10000 );  // No timeout

            // Wait until have signal...
            Console.WriteLine( "Waiting (10s) for signal input...." );
            haveSignal = false;
            while ( !haveSignal )
            {
                Thread.Sleep( 10 );
            }

            if ( irPacket != null )
            {
                var fileName = argParser.GetArgOption( ArgumentParser.CaptureArg );
                if ( IsXmlFile( fileName ) )
                {
                    RRUtil.SerializePacketToXML( fileName, irPacket );
                    Console.WriteLine( "Saved in XML format to file: " + fileName );
                }
                else
                {
                    RRUtil.SerializePacketToBinary( fileName, irPacket );
                    Console.WriteLine( "Saved in binary format to file: " + fileName );
                }
            }
        }

        IRPacket LoadXmlSignal( string fileName )
        {
            var device = argParser.HaveArg( ArgumentParser.DeviceArg ) ? argParser.GetArgOption( ArgumentParser.DeviceArg ) : null;
            var signal = argParser.HaveArg( ArgumentParser.SignalArg ) ? argParser.GetArgOption( ArgumentParser.SignalArg ) : null;
            
            string message;
            var result = XmlSignalLoader.LoadSignal( fileName, device, signal, out message );
            
            if ( result == null )
            {
                Console.WriteLine( message );
            }

            return result;
        }

        protected void OutputSignal()
        {
            var fileName = argParser.GetArgOption( ArgumentParser.OutputArg );
            if ( IsXmlFile( fileName ) )
            {
                irPacket = LoadXmlSignal( fileName );
                if ( irPacket == null ) return;
            }
            else
            {
                irPacket = RRUtil.DeserializePacketFromBinary( fileName );
            }

            // Print signal if requested
            if ( argParser.HavePrintArg )
            {
                Console.WriteLine( irPacket );
            }

            // Output signal
            if ( irPacket is ModulatedSignal )
            {
                redRat.OutputModulatedSignal( irPacket );
                Console.WriteLine( "Output modulated signal." );
            }
            else if ( irPacket is DoubleSignal )
            {
                // Only take first signal
                var sig1 = ( (DoubleSignal)irPacket ).Signal1;
                if ( sig1 != null )
                {
                    redRat.OutputModulatedSignal( sig1 );
                    Console.WriteLine( "Output modulated signal (sig1)." );
                }
            }
            else if ( irPacket is IrDaPacket )
            {
                redRat.OutputIrDaPacket( (IrDaPacket)irPacket );
                Console.WriteLine( "Output IrDa signal." );
            }
        }

        protected void PrintUsage( string errorMessage )
        {
            var headerLine = new StringBuilder();
            // Get version info.
            var assembly = Assembly.GetExecutingAssembly();

            // Title
            var ata = (AssemblyTitleAttribute[])assembly.GetCustomAttributes( typeof( AssemblyTitleAttribute ), true );
            if ( ata.Length > 0 )
            {
                // Take the first one. Do we ever have more???
                headerLine.Append( ata[0].Title + "\n" );
            }

            // Version
            var version = assembly.GetName().Version;
            headerLine.Append( "Version: " + version.Major + "." + version.Minor );
            headerLine.Append( " - Command line interface to the RedRat." );

            Console.WriteLine( headerLine.ToString() );
            Console.WriteLine( "----------------------------------------------------\n" );
            Console.WriteLine( errorMessage );
            Console.WriteLine( "" );
            Console.WriteLine( "  Usage: RR_CLI <RedRat3 name> <options>\n" );
            Console.WriteLine( "  Options are:" );
            Console.WriteLine( "     -blink              Blinks the red LED." );
            Console.WriteLine( "     -capture filename   Captures signal and stores in file <filename>." );
            Console.WriteLine( "     -output filename    Outputs a signal stored in file <filename>." );
            Console.WriteLine( "     -print              Prints the IR signal data when output." );
            Console.WriteLine( "     -device devicename  Specifies the device to use when working with an xml" );
            Console.WriteLine( "                         signal database. (Use -output to set the file.)" );
            Console.WriteLine( "     -signal signalname  Specifies the signal to use when working with an xml" );
            Console.WriteLine( "                         signal database or xml device file. (Use -output to" );
            Console.WriteLine( "                         set the file.)" );
            Console.WriteLine( "" );
            Console.WriteLine( "Press RETURN to continue..." );
            Console.ReadLine();
        }

        /// <summary>
        /// Handles input signal (or error) from the RedRat.
        /// </summary>
        public void SignalDataHandler( object sender, EventArgs e )
        {

            if ( e is SignalEventArgs siea )
            {
                switch ( siea.Action )
                {
                    case SignalEventAction.EXCEPTION:
                        Console.WriteLine( siea.Exception.Message );
                        break;

                    case SignalEventAction.MODULATED_SIGNAL:
                        Console.WriteLine( "Have IR data..." );
                        irPacket = siea.ModulatedSignal;
                        break;

                    case SignalEventAction.IRDA_PACKET:
                        Console.WriteLine( "Have IR data..." );
                        irPacket = siea.IrDaPacket;
                        break;

                    default:
                        Console.WriteLine( "Unknown return..." );
                        break;
                }
                haveSignal = true;
            }
            else
            {
                Console.WriteLine( "Event of unknown type...." );
            }
        }

        // ==== Serializer Handler Interface ====
        public void UnknownNode( object sender, XmlNodeEventArgs e )
        {
            Console.WriteLine( "Error parsing file - unknown Node: " + e.Name + "\t" + e.Text );
        }

        public void UnknownAttribute( object sender, XmlAttributeEventArgs e )
        {
            System.Xml.XmlAttribute attr = e.Attr;
            Console.WriteLine( "Error parsing file - unknown attribute " + attr.Name + "='" + attr.Value + "'" );
        }
    }
}
