﻿using System;
using System.IO;
using System.Linq;
using System.Threading;
using System.Xml.Serialization;
using RedRat.AVDeviceMngmt;
using RedRat.IR;
using RedRat.RedRat4;
using RedRat.RedRat4.StreamIO.IP;
using RedRat.RedRat4.StreamIO.X;
using RedRat.USB;
using RedRat.Util;

namespace RedRat.Samples.RedRatX_Demo
{
    /// <summary>
    /// This sample shows how to use most aspects of the RedRat-X's API. It includes:
    ///     - Discovery and connection
    ///     - Simple operations: Obtaining the firmware version and blinking LEDs.
    ///     - IR signal output via the front blaster and rear IR ports.
    ///     - IR signal input/capture from both the wide-band (learning) detector and the 
    ///       narrow band (standard RC) detectors.
    /// </summary>
    public class RedRatXDemo
    {
        readonly Random rand = new Random();
        public const string dbFilename = "DeviceDB.xml";
        private AVDeviceDB deviceDB;

        static void Main( string[] args )
        {
            try
            {
                new RedRatXDemo( args );
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( ex );
            }
            finally
            {
                Console.WriteLine( "Hit <RETURN> to exit..." );
                Console.ReadLine();
            }
        }

        public RedRatXDemo( string[] args )
        {
            Console.WriteLine( "\nRedRatX Demo\n============\n" );

            // Find any RedRatXs
            var locationInfoList = FindRedRatXs();
            if ( locationInfoList.Length == 0 )
            {
                Console.WriteLine( "No RedRat-X devices found." );
                return;
            }

            // Print out a bit of information about them.
            foreach ( var locInfo in locationInfoList )
            {
                PrintRrxInfo( locInfo );
            }

            // Get actual RedRatX objects to work with. Use the first in the list
            var rrxLi = locationInfoList[0];
            using ( var rrx = RedRatX.GetInstance( rrxLi ) )
            {
                if ( rrx == null )
                {
                    Console.Error.WriteLine( "LocationInfo object was not for a RedRat-X." );
                    return;
                }

                rrx.Connect();

                // Use the RRX a bit
                Console.WriteLine( $"Using RedRat-X '{rrxLi.Name}'" );
                Console.WriteLine( $"Firmware version: {rrx.FirmwareVersion}\n" );

                // Blink the lights a few times
                Console.WriteLine( "Blinking the LEDs..." );
                for ( var i = 0; i < 5; i++ )
                {
                    rrx.Blink();
                    Thread.Sleep( 750 );
                }

                DoIrSignalOutput( rrx );

                // Test IR signal input from both the IR detectors. There are:
                //     1) Wide band detector for IR signal capture/learning (short range)
                //     2) Narrow band, for control of the PC or IR key press logging etc. (long range)
                DoWideBandIrSignalInput( rrx );
                DoNarrowBandIrSignalInput( rrx );

                rrx.Disconnect();

                // If the RR-X is a USB device, test the USB unplug/plug handling.
                if ( LocationInfo.IsUSB( rrx.LocationInformation.RRType ) )
                {
                    TestUsbPlugHandling( rrx );
                }
            }
        }

        /// <summary>
        /// Demo some IR signal output via both the SYNC and ASYC APIs.
        /// </summary>
        private void DoIrSignalOutput( IRedRatX rrx )
        {
            Console.WriteLine( "\n==> IR signal output" );

            // == Sync IR Signal Output ==

            // The RRX has three IR output ports at the back (ports 1, 2 & 3) and a blaster on the front (port 4).
            // With SYNC IR output methods, if no output is enabled then the front blaster is used.

            // Output an IR signal via the blaster
            rrx.OutputModulatedSignal( GetSignal() );
            Console.WriteLine( "IR out via blaster using sync API" );

            // Output an IR signal via port 2 at 40% power
            rrx.EnableIROutput( 2, 40 );
            rrx.OutputModulatedSignal( GetSignal() );
            rrx.DisableIROutput( 2 );
            Console.WriteLine( "IR out via port 2 using sync API" );

            // == Async IR Signal Output ==

            // Using the ASYNC API, it is necessary to enable all ports required, even the front blaster.
            var port = RedRatX.IR_BLASTER_PORT;
            var opList = new OutputList {new OutputPort( port, 30 )};
            var mutex = new ManualResetEvent( false );

            for ( var i = 0; i < 5; i++ )
            {
                rrx.OutputIRPacketAsync( GetSignal(), opList, 100, null, ar =>
                {
                    Console.WriteLine( $"IR out via port {port} using async API." );
                    mutex.Set();
                } );

                mutex.WaitOne(); // Wait for IR output completion
                Thread.Sleep( 750 ); // Wait a bit longer
            }
        }

        /// <summary>
        /// Input an IR signal using the wide band detector for IR signal capture/learning (short range)
        /// </summary>
        private void DoWideBandIrSignalInput( IRedRatX rrx )
        {
            Console.WriteLine( "\n==> IR signal input from wide-band receiver" );
            Console.WriteLine( "Point remote at RR-X and press a button (timeout = 10s)..." );

            var mutex = new ManualResetEvent( false );
            EventHandler learningSignalInHandler = ( s, ea ) =>
            {
                var siea = ea as SignalEventArgs;
                if ( siea == null ) return;
                if ( siea.Action == SignalEventAction.MODULATED_SIGNAL )
                {
                    Console.WriteLine( $"Have IR signal from wide-band detector:\n{siea.ModulatedSignal}" );
                }
                else if ( siea.Action == SignalEventAction.INPUT_CANCELLED )
                {
                    Console.WriteLine( "Timeout when waiting for IR signal." );
                }

                mutex.Set();
            };

            // Hook up the incoming event handler just for this operation. An alternative is to hook it
            // up for the duration of the application.
            rrx.LearningSignalIn += learningSignalInHandler;

            // Set the RR-X up for IR signal input and allow 10s for the RCU button press.
            rrx.GetModulatedSignal( 10000 );
            mutex.WaitOne();
            rrx.LearningSignalIn -= learningSignalInHandler;
        }

        /// <summary>
        /// Use the narrow-band, long-range detector for capturing RCU button presses.
        /// </summary>
        private void DoNarrowBandIrSignalInput( IRedRatX rrx )
        {
            Console.WriteLine( "\n==> IR signal input from narrow-band receiver" );
            Console.WriteLine( "Point remote at RR-X and press a few buttons..." );

            EventHandler rcDetectorSignalInHandler = ( s, ea ) =>
            {
                var siea = ea as SignalEventArgs;
                if ( siea == null ) return;
                if ( siea.Action == SignalEventAction.MODULATED_SIGNAL )
                {
                    Console.WriteLine( $"Have IR signal from RC detector:\n{siea.ModulatedSignal}" );
                    Console.WriteLine( "Hit <RETURN> to stop RC signal input." );
                }
            };

            // Hook up the incoming event handler just for this operation. An alternative is to hook it
            // up for the duration of the application.
            rrx.RCDetectorSignalIn += rcDetectorSignalInHandler;

            // Start the RR-X listening and then continue to input IR signals until <RETURN> is hit.
            rrx.RCDetectorEnabled = true;
            Console.WriteLine( "Hit <RETURN> to stop RC signal input." );
            Console.ReadLine();

            rrx.RCDetectorEnabled = false;
            rrx.RCDetectorSignalIn -= rcDetectorSignalInHandler;
            Console.WriteLine( "RC signal input terminated.\n" );
        }

        /// <summary>
        /// Hook up to USB plug events and print out details as they happen.
        /// </summary>
        private void TestUsbPlugHandling( IRedRatX rrx )
        {
            // Create local plug/unplug handler
            EventHandler<UsbPlugHandler.UsbPlugHandlerEventArgs> usbPlugHandler = ( sender, args ) =>
            {
                Console.WriteLine( "Plug event... " + args );
                if ( args.plugEventType == UsbPlugHandler.DevicePlugEventEnum.Added )
                {
                    // The plug event gives us the serial number of the device, so use this to 
                    // find the RR-X that has just been plugged in.
                    var usbLis = RRUtil.FindRedRats( LocationInfo.RedRatType.RedRatX_USB ).OfType<RedRatXUsbLocationInfo>();
                    var usbLi = usbLis.FirstOrDefault( li => li.SerialNumber == args.deviceSerialNumber );

                    Console.WriteLine( $"Plugged in device is '{usbLi.Name}'" );
                }
            };

            // Add plug/unplug handler
            UsbPlugHandler.Instance.UsbPlugEvent += usbPlugHandler;

            Console.WriteLine( "Unplug + replug USB RedRat-X and then hit <RETURN>" );
            Console.ReadLine();

            // Remove plug handler
            UsbPlugHandler.Instance.UsbPlugEvent -= usbPlugHandler;
        }

        /// <summary>
        /// Obtain a list of objects describing information about any RedRat-Xs found.
        /// </summary>
        /// <param name="usbOnly">If set to true, only RedRatXs connected via USB will be returned, otherwise both USB and network
        /// connected devices will be listed. A USB search is almost instantaneous but a network search takes a few seconds.</param>
        /// <returns>An array of LocationInfo objects; one for each RR-X discovered.</returns>
        private LocationInfo[] FindRedRatXs( bool usbOnly = false )
        {
            if ( !usbOnly )
            {
                Console.WriteLine( "Searching for RedRat-X devices." );
            }

            var searchType = usbOnly ? LocationInfo.RedRatType.RedRatX_USB : LocationInfo.RedRatType.RedRatX;
            return RRUtil.FindRedRats( searchType );
        }

        /// <summary>
        /// Prints out information about the RedRatX from the passed LocationInfo object.
        /// </summary>
        private void PrintRrxInfo( LocationInfo locInfo )
        {
            Console.WriteLine( $"RedRat device: {locInfo}" );

            // Due to the differences in the underlying comm mechanism, there are two types of LocationInfo for the RedRat-X:
            //      1) RedRatXIpLocationInfo for IP connected RRXs.
            //      2) RedRatXUsbLocationInfo for USB connected RRXs.
            if ( locInfo is RedRatXIpLocationInfo ipLocInfo )
            {
                Console.WriteLine( $"  Serial#: {ipLocInfo.SerialNumber}" );
                Console.WriteLine( $"       IP: {ipLocInfo.IPAddress}" );
                Console.WriteLine( $"      MAC: {IpUtils.MacAddrBytesToStr( ipLocInfo.MACAddress )}\n" );
            }

            if ( locInfo is RedRatXUsbLocationInfo usbLocInfo )
            {
                Console.WriteLine( $"   Serial#: {usbLocInfo.SerialNumber}\n" );
            }
        }

        /// <summary>
        /// Returns a random IR signal for output.
        /// </summary>
        public ModulatedSignal GetSignal()
        {
            // Load device DB containing IR signals if have not already done so.
            if ( deviceDB == null )
            {
                FileStream fs = null;
                try
                {
                    var ser = new XmlSerializer( typeof( AVDeviceDB ) );
                    fs = new FileStream( new FileInfo( dbFilename ).FullName, FileMode.Open );
                    deviceDB = (AVDeviceDB) ser.Deserialize( fs );
                }
                finally
                {
                    fs?.Close();
                }
            }

            // Get a random IR signal
            ModulatedSignal signal = null;
            // Not all IR signals read from the DB are modulated signals, so loop until we get one.
            while ( signal == null )
            {
                var index = rand.Next( deviceDB.AVDevices.Length );
                var device = deviceDB.AVDevices[index];
                signal = device.Signals[rand.Next( device.Signals.Length )] as ModulatedSignal;
            }

            return signal;
        }

    }
}
