using System;
using System.Drawing;
using System.Windows.Forms;
using System.IO;
using System.Xml.Serialization;

using RedRat.RedRat3;
using RedRat.RedRat3.USB;
using RedRat.AVDeviceMngmt;

namespace RedRat.Samples.SignalDecoding
{
    /// <summary>
    /// Shows how to use the decoding of IR signal to move a windows around the screen.
    /// 
    /// Once started, right-click on the background to load a signal DB that contains the codes
    /// Up, Down, Left and Right for your remote control.
    /// </summary>
    public class SignalDecoder : Form
    {
        /// <summary>
        /// Describes which direction the window is moving in.
        /// </summary>
        public enum Direction
        {
            UP, DOWN, LEFT, RIGHT, NONE
        }

        /// <summary>
        /// Initial step size for window movement.
        /// </summary>
        public const int INIT_STEP_SIZE = 5;

        /// <summary>
        /// The database of signals we use for signal decoding.
        /// </summary>
        private AVDeviceDB avDeviceDB;

        /// <summary>
        /// The actual RedRat3 we're using...
        /// </summary>

        protected IRedRat3 redRat3;

        /// <summary>
        /// Step size use to move the window.
        /// </summary>
        protected double stepSize = INIT_STEP_SIZE;

        /// <summary>
        /// Direction of last window move.
        /// </summary>
        protected Direction lastMove = Direction.NONE;

        /// <summary>
        /// Time that the last move was made.
        /// </summary>
        protected DateTime lastMoveTime = DateTime.Now;

        /// <summary>
        /// Max time between button presses before the step size increase is reset.
        /// Units are "ticks", i.e. 100ns intervals.
        /// </summary>
        protected TimeSpan maxTimeDiff = new TimeSpan( 2000000 );

        /// <summary>
        /// Size of screen.
        /// </summary>
        protected Rectangle workingScreenSize;

        private ContextMenu contextMenu;
        private MenuItem openDBMenuItem;
        private MenuItem exitMenuItem;

        /// <summary>
        /// Required designer variable.
        /// </summary>
        private readonly System.ComponentModel.Container components;

        public SignalDecoder()
        {
            //
            // Required for Windows Form Designer support
            //
            InitializeComponent();
            workingScreenSize = Screen.GetWorkingArea( this );

            OpenRedRat3();

            // Setup for IR signal input from the remote control.
            if ( redRat3 != null )
            {
                // Add listener for incoming signals.
                redRat3.RCDetectorSignalIn += RcSignalDataIn;
                redRat3.RCDetectorEnabled = true;
            }
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        protected override void Dispose( bool disposing )
        {
            // Remove ourselves as listener...
            if ( redRat3 != null )
            {
                redRat3.RCDetectorSignalIn -= RcSignalDataIn;
                redRat3.RCDetectorEnabled = false;
            }

            if ( disposing )
            {
                components?.Dispose();
            }
            base.Dispose( disposing );

            // Close RR3s last of all.
            RedRat3USBImpl.DisposeOfAll();
        }

        #region Windows Form Designer generated code
        /// <summary>
        /// Required method for Designer support - do not modify
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(SignalDecoder));
            this.contextMenu = new System.Windows.Forms.ContextMenu();
            this.openDBMenuItem = new System.Windows.Forms.MenuItem();
            this.exitMenuItem = new System.Windows.Forms.MenuItem();
            this.SuspendLayout();
            // 
            // contextMenu
            // 
            this.contextMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.openDBMenuItem,
            this.exitMenuItem});
            // 
            // OpenDBMenuItem
            // 
            this.openDBMenuItem.Index = 0;
            this.openDBMenuItem.Text = "Open Signal DB";
            this.openDBMenuItem.Click += new System.EventHandler(this.OpenDBMenuItem_Click);
            // 
            // ExitMenuItem
            // 
            this.exitMenuItem.Index = 1;
            this.exitMenuItem.Text = "Exit";
            this.exitMenuItem.Click += new System.EventHandler(this.ExitItem_Click);
            // 
            // SignalDecoder
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.BackgroundImage = ((System.Drawing.Image)(resources.GetObject("$this.BackgroundImage")));
            this.ClientSize = new System.Drawing.Size(320, 152);
            this.ContextMenu = this.contextMenu;
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedToolWindow;
            this.Name = "SignalDecoder";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
            this.Text = " Please load a signal database";
            this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.SignalDecoderFormClosing);
            this.ResumeLayout(false);

        }
        #endregion

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            Application.Run( new SignalDecoder() );
        }

        /// <summary>
        /// Opens the first RedRat3 found.
        /// </summary>
        protected void OpenRedRat3()
        {
            try
            {
                // Find the no. of RR3s connected.
                var devices = RedRat3USBImpl.FindDevices();
                if ( devices.Count > 0 )
                {
                    // Just take the first device found.
                    redRat3 = (IRedRat3)devices[0].GetRedRat();
                }
                else
                {
                    MessageBox.Show( "No RedRat3 devices found.", "TestRemote Warning",
                      MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                }

            }
            catch ( Exception ex )
            {
                MessageBox.Show( ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
            }
        }

        /// <summary>
        /// Move window depending on the returned signal.
        /// </summary>
        protected void MoveWindow( SignalKey sigKey )
        {
            // Read current location
            var curLoc = Location;

            // Should not get a null return, but check in case.
            if ( sigKey == null || sigKey.AVDevice == null )
            {
                lastMove = Direction.NONE;
                stepSize = 1;
                return;
            }

            // If we only want to respond to buttons on remote "Sky Digibox" then uncomment this line.
            // Otherwise respond to any remote we know that has UP, DOWN, LEFT and RIGHT buttons.
            // if ( !sigKey.AVDevice.Name.Equals( "Sky Digibox" ) ) return;

            if ( sigKey.Signal.Name.Equals( "Down", StringComparison.InvariantCultureIgnoreCase ) )
            {
                // Move down
                AdjustStepSize( Direction.DOWN );
                curLoc.Y += (int) stepSize;

                // If window moves off screen, move back onto other side
                if ( curLoc.Y > workingScreenSize.Height )
                {
                    curLoc.Y = 0;
                }

            }
            else if ( sigKey.Signal.Name.Equals( "Up", StringComparison.InvariantCultureIgnoreCase ) )
            {
                // Move up
                AdjustStepSize( Direction.UP );
                curLoc.Y -= (int) stepSize;

                // If window moves completely off screen, move back onto other side
                if ( curLoc.Y + Height < 0 )
                {
                    curLoc.Y = ( workingScreenSize.Height - Height );
                }

            }
            else if ( sigKey.Signal.Name.Equals( "Right", StringComparison.InvariantCultureIgnoreCase ) )
            {
                // Move right
                AdjustStepSize( Direction.RIGHT );
                curLoc.X += (int) stepSize;

                // If window moves off screen, move back onto other side
                if ( curLoc.X > workingScreenSize.Width )
                {
                    curLoc.X = 0;
                }

            }
            else if ( sigKey.Signal.Name.Equals( "Left", StringComparison.InvariantCultureIgnoreCase ) )
            {
                // Move left
                AdjustStepSize( Direction.LEFT );
                curLoc.X -= (int) stepSize;

                // If window moves off screen, move back onto other side
                if ( curLoc.X + Width < 0 )
                {
                    curLoc.X = ( workingScreenSize.Width - Width );
                }

            }
            else
            {
                lastMove = Direction.NONE;
                stepSize = 1;
            }

            SetLocation( curLoc );
        }

        /// <summary>
        /// Sets the form to the new location.
        /// </summary>
        protected void SetLocation( Point newLocation )
        {
            // Possibly (actually very likely) to be a cross-thread GUI op.
            if ( InvokeRequired )
            {
                Invoke( new Action( () => Location = newLocation ) );
            }
            else
            {
                Location = newLocation;
            }
        }

        /// <summary>
        /// Adjusts the speed at which we move given repeated button press.
        /// </summary>
        protected void AdjustStepSize( Direction thisMove )
        {
            var now = DateTime.Now;
            if ( ( lastMove == thisMove ) && ( now - lastMoveTime < maxTimeDiff ) )
            {
                if ( stepSize < 45 )
                {
                    // Only increase if haven't reached max speed.
                    stepSize *= 1.2;
                }
            }
            else
            {
                // Reset
                stepSize = INIT_STEP_SIZE;
            }
            // Store for next time round
            lastMove = thisMove;
            lastMoveTime = now;
        }

        /// <summary>
        /// Load a signal DB for use with decoding.
        /// </summary>
        private void OpenDBMenuItem_Click( object sender, EventArgs e )
        {

            // Read signal database from XML file.
            var openFileDialog = new OpenFileDialog
                                     {
                                         Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*",
                                         RestoreDirectory = true
                                     };

            if ( openFileDialog.ShowDialog() != DialogResult.OK ) return;

            var fName = openFileDialog.FileName;
            var ser = new XmlSerializer( typeof ( AVDeviceDB ) );
            FileStream fs = null;
            try
            {
                fs = new FileStream( ( new FileInfo( fName ) ).FullName, FileMode.Open );
                var newAVDeviceDB = (AVDeviceDB) ser.Deserialize( fs );
                avDeviceDB = newAVDeviceDB;
                Text = "";
            }
            catch ( Exception ex )
            {
                MessageBox.Show( ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
            }
            finally
            {
                fs?.Close();
            }
        }

        private void ExitItem_Click( object sender, EventArgs e )
        {
            Close();
        }


        /// <summary>
        /// Handles input signal events from RedRat3 RC detector (or error).
        /// This is minimum code, i.e. does not have good error handling etc.
        /// </summary>
        public void RcSignalDataIn( object sender, EventArgs e )
        {
            if ( !( e is SignalEventArgs siea ) ) return;         // Who called this then???

            switch ( siea.Action )
            {
                case SignalEventAction.EXCEPTION:
                    MessageBox.Show( siea.Exception.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                    break;

                case SignalEventAction.MODULATED_SIGNAL:
                    if ( avDeviceDB != null )
                    {
                        // Decode signal if have DB.
                        try
                        {
                            var sigKey = avDeviceDB.DecodeSignal( siea.ModulatedSignal );
                            MoveWindow( sigKey );
                        }
                        catch ( Exception ex )
                        {
                            Console.WriteLine( ex.Message + "\n" + ex.StackTrace + "\n" );
                        }
                    }
                    break;

                default:
                    break;
            }
        }

        private void SignalDecoderFormClosing( object sender, FormClosingEventArgs e )
        {
            // Try and ensure we don't get the RR3 hanging.
            if ( redRat3 != null )
            {
                redRat3.RCDetectorEnabled = false;
            }
        }
    }
}
