﻿using System;
using System.IO;
using System.Linq;
using System.Xml.Serialization;

using RedRat.IR;
using RedRat.RedRat3;
using RedRat.Util;
using RedRat.AVDeviceMngmt;
using RedRat.RedRat3.USB;

namespace RedRat.Samples.SimpleSignalOutput
{
    /// <summary>
    /// This sample shows how to read in a signal database XML file, extract one signal from it
    /// and then output via a RedRat3 device.
    /// 
    /// The database file, device and signal names are given as command line parameters:
    /// 
    ///     SimpleSignalOutput database-file.xml device-name signal-name
    /// e.g. 
    ///     SignalSignalOutput DeviceDB.xml CD Play
    /// 
    /// It is not an efficient way of sending sequences of IR signals as loading the XML file
    /// for each signal is quite an expensive operation.
    /// </summary>
    public class SimpleSignalOutput
    {
        static void Main(string[] args)
        {
            new SimpleSignalOutput( args );
        }

        public SimpleSignalOutput( string[] args )
        {
            try
            {
                CheckArgs( args );
                using ( var rr3 = FindRedRat3() )
                {
                    var signalDB = LoadSignalDB( args[0] );
                    var signal = GetSignal( signalDB, args[1], args[2] );
                    rr3.OutputModulatedSignal( signal );
                    Console.WriteLine( $"Signal {args[1]}->{args[2]} output.\n" );
                }
            }
            catch ( Exception ex )
            {
                Console.WriteLine( "Error: " + ex.Message );
            }
        }

        /// <summary>
        /// Simple check of program arguments. We need all three in the correct order to work.
        /// </summary>
        private void CheckArgs( string[] args )
        {
            // Need all three arguments.
            if ( ( args != null ) && ( args.Length == 3 ) ) return;
            throw new Exception( "Invalid argument list.\n\nUsage: SimpleSignalOutput database-file.xml device-name signal-name\n");
        }

        /// <summary>
        /// Simply finds the first RedRat3 attached to this computer.
        /// </summary>
        private IRedRat3 FindRedRat3()
        {
            var rr3li = RRUtil.FindRedRats( LocationInfo.RedRatType.RedRat3 ).FirstOrDefault();

            if ( rr3li == null )
            {
                throw new Exception( "Unable to find any RedRat3 devices on this computer." );
            }
            return rr3li.GetRedRat() as IRedRat3;;
        }

        /// <summary>
        /// Loads the signal database XML file.
        /// </summary>
        private AVDeviceDB LoadSignalDB( string dbFileName )
        {
            var ser = new XmlSerializer( typeof ( AVDeviceDB ) );
            var fs = new FileStream( new FileInfo( dbFileName ).FullName, FileMode.Open );
            var avDeviceDB = (AVDeviceDB) ser.Deserialize( fs );
            return avDeviceDB;
        }

        /// <summary>
        /// Returns an IR signal object from the signal DB file using the deviceName and signalName to
        /// look it up.
        /// </summary>
        private IRPacket GetSignal( AVDeviceDB signalDB, string deviceName, string signalName )
        {
            var device = signalDB.GetAVDevice( deviceName );
            if ( device == null )
            {
                throw new Exception( $"No device of name '{deviceName}' found in the signal database." );
            }
            var signal = device.GetSignal( signalName );
            if ( signal == null )
            {
                throw new Exception( $"No signal of name '{signalName}' found for device '{deviceName}' in the signal database." );
            }
            return signal;
        }
    }
}
