//
// This is a simple demo application in VC++.NET for the RedRat3.
// Managed extensions are used.
//
// To get this project to compile and run with the RedRat assemblies (DLLs), 
// a couple of project settings may to be changed, depending on your setup.
// This is done in the project property pages (right click on project and
// select "Properties"):
//
// 1) "C/C++->General->Resolve #using References"
//    Add the directory in which the RedRat.dll is placed.
//
// 2) The RedRat.dll should be in the same directory as the exe generated from 
//    this project (unless you are installing it in the GAC) and so a post build 
//    action has to be setup to copy them to the output dir:
//
//    "Build Events->Post-Build Event->Command Line"
//    Add a command to copy the dlls from wherever they are placed to the 
//    project target directory, e.g.
//           copy "$(ProjectDir)..\dependencies\*.dll" "$(TargetDir)"
//

#include "stdafx.h"

#using <mscorlib.dll>

#include <tchar.h>

using namespace System;
using namespace System::Collections::Generic;
using namespace System::IO;
using namespace System::Xml::Serialization;

using namespace RedRat;
using namespace RedRat::IR;
using namespace RedRat::USB;
using namespace RedRat::Util;
using namespace RedRat::RedRat3;
using namespace RedRat::RedRat3::USB;
using namespace RedRat::AVDeviceMngmt;

//
// Event handler class for remote control input using the
// long range remote control detector on RedRat.
//
[event_receiver(managed)]
ref class CRCDataIn {
public:
    void SignalIn(Object^ sender, EventArgs^ e) {
        // The event sent from the RR3 could mean one of three things:
        // i)   We have an input IR signal
        // ii)  The signal input has been cancelled (send via exception)
        // iii) There has been some error in signal input (also an exception)

        // Right - this is very C# like but looks a lot worse. I'm sure there's a more
        // elegant way of doing this in managed VC++??
        if (dynamic_cast<SignalEventArgs^>(e) != nullptr) {
            SignalEventArgs^ siea = static_cast<SignalEventArgs^>(e);

            // Exception
            if (siea->Action == SignalEventAction::EXCEPTION) {
                // Means that we have had an error....
                Console::WriteLine(String::Concat("EVENT --> Exception...", siea->Exception->ToString()));
            } 

            // Input signal
            else if (siea->Action == SignalEventAction::MODULATED_SIGNAL) {
                Console::WriteLine("EVENT --> Have signal.");
            } 

            // RC detector enabled.
            else if (siea->Action == SignalEventAction::RC_DETECTOR_ENABLED) {
                Console::WriteLine("EVENT --> RC detector enabled.");
            } 

            // RC detector disabled.
            else if (siea->Action == SignalEventAction::RC_DETECTOR_DISABLED) {
                Console::WriteLine("EVENT --> RC detector disabled.");
            }
        }
    }

    void HookEvent(IRedRat3^ rr3) {
        rr3->RCDetectorSignalIn += gcnew EventHandler(this, &CRCDataIn::SignalIn);
    }

    void UnhookEvent(IRedRat3^ rr3) {
        rr3->RCDetectorSignalIn += gcnew EventHandler(this, &CRCDataIn::SignalIn);
    }
};


int _tmain(void)
{
    Console::WriteLine("\nVC++.NET RedRat3 demo program\n");

    try {
        // Find the list of RedRat3s attached to this system.
		IList<LocationInfo^>^ rr3Li = RedRat3USBImpl::FindDevices();

        // If no RedRats, then quit.
        if (rr3Li->Count == 0) {
            Console::WriteLine("No RedRats found.");
            return 0;
        }

        // Print out names
        for (int i=0; i<rr3Li->Count; i++) {
            Console::WriteLine(String::Concat("   Found ", rr3Li[i]->Name));
        }

        // Take the first RR3 and use that.
        IRedRat3^ redRat3 = (IRedRat3^)rr3Li[0]->GetRedRat();

        // == Blink the LED ==
        redRat3->Blink();
        Console::WriteLine("--> Blinked");

        // == Show some device information ==
        USBDeviceInfo^ devInfo = static_cast<USBDeviceInfo^>(redRat3->DeviceInformation);
        Console::WriteLine("\nDevice information:");
        Console::WriteLine(String::Concat("  Company     : ", devInfo->Company));
        Console::WriteLine(String::Concat("  Product Name: ", devInfo->ProductName));
		Console::WriteLine(String::Concat("  Version     : ", devInfo->ProductDescriptor->Version->ToString()));
        Console::WriteLine(String::Concat("  Serial No.  : ", devInfo->SerialNumber, "\n"));

        // == Learn a signal ==
        // 1. Create and hookup the event handler
        CRCDataIn^ sigDataHandler = gcnew CRCDataIn();
        sigDataHandler->HookEvent(redRat3);

        // 2. Enable learning detector and inform user.
        Console::Write("Waiting for signal from RC detector on RedRat...");
        Console::WriteLine("press return to terminate.");
        redRat3->RCDetectorEnabled = true;

        // 3. Wait until we've had enough
        Console::ReadLine();

        // 4. Disable signal input (important) and Unhook event handlers
        redRat3->RCDetectorEnabled = false;
        sigDataHandler->UnhookEvent(redRat3);

        // == Try loading a signal DB, listing contents and outputting a signal ==
        FileStream^ fs; 
        try {
            // Load file
            XmlSerializer^ ser = gcnew XmlSerializer(AVDeviceDB::typeid);
            fs = gcnew FileStream((gcnew FileInfo("D:\\temp\\DeviceDB.xml"))->FullName, FileMode::Open); 
            AVDeviceDB^ avDeviceDB = static_cast<AVDeviceDB^>(ser->Deserialize(fs)); 
            fs->Close();

            // List devices
            array<AVDevice^>^ avDevices = avDeviceDB->AVDevices;
            for (int i=0; i<avDevices->Length; i++) {
                Console::WriteLine(String::Concat("AVDevice: ", avDevices[i]->Name));
                // List signals
                array<IRPacket^>^ irPackets = avDevices[i]->Signals;
                for (int j=0; j<irPackets->Length; j++) {
                    Console::WriteLine(String::Concat("  +--> signal: ", irPackets[j]->Name));
                }
            }

            // Output a signal
            IRPacket^ sig = avDeviceDB->GetIRPacket("CD", "Play");
            // Check its a modulated signal (is there a better way in C++??)
            if (dynamic_cast<ModulatedSignal^>(sig) != nullptr) {
                ModulatedSignal^ modSig = static_cast<ModulatedSignal^>(sig);
                redRat3->OutputModulatedSignal(modSig);
                Console::WriteLine(String::Concat("\nIR signal output"));
            }

        } catch (Exception^ ex) { 
            Console::WriteLine(String::Concat("Exception opening XML file: ", ex->ToString()));
            if (fs) fs->Close();
        }

    } catch (System::Exception^ ex) {
        Console::WriteLine(String::Concat("Exception: ", ex->ToString()));
    }

    return 0;
}
