﻿using System;
using System.Threading;
using RedRat.IR;
using RedRat.RedRat4;

namespace RedRat.Samples.IrNetBoxDemo
{
    /// <summary>
    /// Outputs IR signals asynchronously using one output. To be run in its  own thread.
    /// </summary>
    public class IrNbIIISignalOutputWorker
    {
        private readonly IRedRat4_III rr4;
        private readonly int port;
        private readonly Func<ModulatedSignal> GetSignal;
        private readonly int opPowerLevel;
        private readonly object opLock = new object();
        private readonly Action ThreadCompleted;
        private readonly int numberToOutput;

        public const int postSigDelay = 100;

        public IrNbIIISignalOutputWorker( IRedRat4_III rr4, int port, Func<ModulatedSignal> getSignal, int powerLevel, 
            int numberToOutput, Action threadCompleted )
        {
            this.rr4 = rr4;
            this.port = port;
            GetSignal = getSignal;
            opPowerLevel = powerLevel;
            this.numberToOutput = numberToOutput;
            ThreadCompleted = threadCompleted;
        }

        /// <summary>
        /// Main loop - sends each IR signal via the one port in turn. 
        /// </summary>
        public void SignalOutputLoop( object state )
        {
            try
            {
                // Setup output list information for the IR output call. In this case it is a single
                // output - the port that this thread is responsible for.
                var opList = new OutputList
                                 {
                                     new OutputPort( port, opPowerLevel ),
                                 };
                var count = 0;
                while ( count < numberToOutput)
                {
                    // Send the IR signal. If this port is busy then an error is returned, so we have to wait
                    // until the previous IR signal output has completed. We are notified of completion in the
                    // IROutputCallback method.
                    rr4.OutputIRPacketAsync( GetSignal(), opList, postSigDelay, count, IROutputCallback );
                    lock ( opLock )
                    {
                        // Wait for IR output completion.
                        Monitor.Wait( opLock );
                    }
                    count++;

                    // Don't really need this, but reduce speed a bit to help see what's going on the box's display.
                    Thread.Sleep( 250 );
                }
            }
            catch( Exception ex )
            {
                Console.WriteLine( ex );
            }
            finally
            {
                // Indicate to main thread that we have finished. This calls the
                // WorkerThreadCompleted method in the main class.
                Console.WriteLine( $"\t>> Worker thread for port {port} complete." );
                ThreadCompleted(); 
            }
        }

        /// <summary>
        /// Callback from RedRat.dll to indicate completion of the IR output.
        /// </summary>
        private void IROutputCallback( IROutputAsyncResult asyncResult )
        {
            lock ( opLock )
            {
                // Give the main output thread a bump so that it can continue and send the next IR signal.
                Monitor.Pulse( opLock );
            }

            // Do we have an error returned?
            if ( asyncResult.Error != null )
            {
                Console.WriteLine( asyncResult.Error );
            }

            var count = (int)asyncResult.State;
        }
    }
}
