﻿using System;
using System.IO;
using System.Linq;
using System.Threading;
using System.Xml.Serialization;
using RedRat.AVDeviceMngmt;
using RedRat.IR;
using RedRat.RedRat4;
using RedRat.RedRat4.StreamIO;
using RedRat.RedRat4.StreamIO.IP;
using RedRat.Util;

namespace RedRat.Samples.IrNetBoxDemo
{
    /// <summary>
    /// Example code that shows how to use the RedRat API to communicate with an irNetBox-III.
    /// 
    /// When compiled, run as: "irNetBoxDemo.exe [IP address | MAC address | name]". For example:
    /// 
    ///    > irNetBoxDemo.exe 192.168.1.40
    ///    > irNetBoxDemo.exe 00-20-4A-CF-2B-82
    ///    > irNetBoxDemo.exe "irNB Rack-1"
    /// 
    /// This code shows three ways of IR output via the irNetBox:
    /// 
    /// 1) Using the old API in an identical fashion to the MK-II box. One IR signal is sent
    ///    at a time and there is no speed improvement.
    /// 
    /// 2) Using the Async Output Manager. This is basically a queue that adds as many IR signals
    ///    as needed and sends them on to the irNetBox-III as quickly as possible. It is simple to
    ///    use and does the clever stuff of ensuring that IR signals are not sent to busy output
    ///    ports. The disadvantage is that it does not give a detailed level of control over the
    ///    irNetBox-III and -IV.
    /// 
    /// 3) Directly using the new async API. This is more complex but gives better control of the
    ///    box. The example here starts one worker thread for each output port and sends IR signals
    ///    to that port as fast as possible.
    /// </summary>
    public class IRNetBoxDemo
    {
        /// <summary>
        /// Use this IP address if none given as first command line args.
        /// </summary>
        public const string defaultIPAddress = "192.168.1.170";

        readonly Random rand = new Random();
        public const string dbFilename = "DeviceDB.xml";
        private AVDeviceDB deviceDB;

        // Variables for thread sync.
        private int threadCount;
        private ManualResetEvent finished;

        static void Main( string[] args )
        {
            try
            {
                new IRNetBoxDemo( args );
            }
            catch ( Exception ex )
            {
                Console.Error.WriteLine( ex );
            }
            finally
            {
                Console.WriteLine( "Hit <RETURN> to exit..." );
                Console.ReadLine();
            }
        }

        /// <summary>
        /// .ctor
        /// </summary>
        public IRNetBoxDemo( string[] args )
        {
            // Explicitly increase the number of available worker threads.
            ThreadPool.SetMinThreads( 20, 20 );

            Console.WriteLine( "\nIRNetBoxDemo\n=================\n" );

            // The first job is to find the irNetBox we want.
            var locationInfo = FindIrNetBox( args.Length > 0 ? args[0].Trim() : defaultIPAddress );
            if ( locationInfo == null )
            {
                return;
            }

            // Get a generic irNetBox (IRedRat4) object to use
            using ( var irNB = locationInfo.GetRedRat() as IRedRat4 )
            {

                // Check that the cast has worked ok. It should have worked fine this case as the original search 
                // was just for these type of RedRat devices, but good practice in case this code is used elsewhere.
                if (irNB == null)
                {
                    Console.WriteLine("The irNetBox with IP/MAC/name of '{0}' is not of type MK-III or MK-IV.", args[0]);
                    return;
                }

                // Connect to irNetBox.
                irNB.Connect();

                // Blink a few times...
                Console.WriteLine("Blinking...");
                for ( var i = 0; i < 4; i++ )
                {
                    irNB.Blink();
                    Thread.Sleep( 500 );
                }

                // ===== Do work ====
                // 1. Using synchronous API. No faster than the MK-II irNetBox.
                DoSyncAPICalls( irNB );

                // Check to see if this is a MK-III (or greater) irNetBox
                if ( LocationInfo.IsMKIII( irNB.LocationInformation.RRType ) || LocationInfo.IsMKIV( irNB.LocationInformation.RRType ) )
                {
                    // Cast as a MK-III as this gives us the async API to use.
                    var irNBIII = irNB as IRedRat4_III;

                    // 2. Using the async output manager. Easy but less control.
                    DoAsyncOutputManagerCalls( irNBIII );

                    // 3. Using the async API. More involved (not much) but more control.
                    DoAsyncAPICalls( irNBIII );
                }

                // Disconnect from box
                irNB.Disconnect();
            }
        }

        /// <summary>
        /// Given an IP address, MAC address or friendly name, find the irNetBox we want.
        /// </summary>
        public LocationInfo FindIrNetBox( string irNetBoxId )
        {
            // If the IP address is dynamically assigned, then we may not know what it is, so we should use:
            //    1) The MAC address as given on the back of the irNetBox.
            //    2) A friendly name, which can be assigned with a RedRat application such as the 
            //       Signal DB Utility (RedRats menu item).

            // Do a network search (broadcast) to find any irNetBoxes out there. This will take a few seconds to return.
            // The search returns a list of LocationInfo objects which are used to provide information about the irNetBoxes.
            // We can cast them all to type RedRat4IpLocationInfo as we are only searching for IP irNetBoxes.
            Console.WriteLine("Searching for irNetBoxes....");
            var irNBLocInfos = RRUtil.FindRedRats( LocationInfo.RedRatType.RedRat4 ).Cast<RedRat4IpLocationInfo>();

            // Print the list...
            if ( !irNBLocInfos.Any() )
            {
                Console.WriteLine("No irNetBoxes found on network.");
                return null;
            }

            Console.WriteLine( "Found the following:");
            foreach (var locationInfo in irNBLocInfos)
            {
                Console.WriteLine( $"\t{locationInfo.IPAddress}" +
                                   $"\t{IpUtils.MacAddrBytesToStr(locationInfo.MACAddress)}" +
                                   $"\t{locationInfo.RRType}\t'{locationInfo.Name}'" );
            }
            Console.WriteLine();

            // Check IP address...
            var irNbLocationInfo = irNBLocInfos.FirstOrDefault( rr => rr.IPAddress.ToString().Equals( irNetBoxId ) );
            if ( irNbLocationInfo != null )
            {
                Console.WriteLine( $"Found irNetBox with IP = {irNbLocationInfo.IPAddress}" );
                return irNbLocationInfo;
            }

            // Check MAC address...
            irNbLocationInfo = irNBLocInfos.FirstOrDefault( rr => IpUtils.MacAddrBytesToStr( rr.MACAddress ).Equals( irNetBoxId ) );
            if (irNbLocationInfo != null)
            {
                Console.WriteLine( $"Found irNetBox with MAC address = '{IpUtils.MacAddrBytesToStr(irNbLocationInfo.MACAddress)}'" );
                return irNbLocationInfo;
            }

            // Check friendly name...
            irNbLocationInfo = irNBLocInfos.FirstOrDefault(rr => rr.Name.Equals( irNetBoxId ) );
            if (irNbLocationInfo != null)
            {
                Console.WriteLine( $"Found irNetBox with name = '{irNbLocationInfo.Name}'" );
                return irNbLocationInfo;
            }

            // Can't find the irNetBox.
            Console.WriteLine( $"No irNetBox found with IP/MAC/Name of {irNetBoxId}" );
            return null;
        }

        /// <summary>
        /// Use the standard synchronous API. This works for all types of irNetBox, but it is quite slow as
        /// it only sends one IR signal at a time. For MK-III (or higher) irNetBoxes, the async API can be used instead.
        /// </summary>
        public void DoSyncAPICalls( IRedRat4 irNB )
        {
            Console.WriteLine( "\nSYNC API CALLS" );

            // Output an IR signal via a few enabled outputs
            Console.WriteLine( "\tOutputting a couple of IR signals..." );
            var opState = new OutputList
            {
                new OutputPort( 2, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) ),
                new OutputPort( 4, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) ),
                new OutputPort( 5, OutputListUtil.OutputPowerToInt( OutputPower.LOW ) )
            };
            irNB.EnabledIROutputs = opState; // Set the output state.
            for ( var i = 0; i < 4; i++ )
            {
                // Send the IR signal to the enabled outputs.
                irNB.OutputModulatedSignal( GetSignal() );
                Thread.Sleep( 500 );
            }
            Console.WriteLine( "\tDone." );
        }

        /// <summary>
        /// Output IR signals using the async output manager. This is basically a queue into which IR
        /// signals are inserted and they are output as quickly as possible. The disadvantage is that
        /// one doesn't get notified when a particular IR signal output is complete - for that one
        /// needs the normal async API.
        /// </summary>
        public void DoAsyncOutputManagerCalls( IRedRat4_III irNBIII )
        {
            Console.WriteLine( "\nASYNC OUTPUT MANAGER CALLS" );

            // Display info if MK-IV
            var irNBIV = irNBIII as IRedRat4_IV;
            irNBIV?.DisplayText( "Async o/p V1" );

            // Get the Async Output Manager.
            var asynOutputManager = ( irNBIII as RedRat4StreamIOImpl ).AsyncOutputManager;

            const int numberOfSignals = 200; // Output this many IR signals.

            // Send a load of IR signals to a random selection of ports.
            Console.WriteLine( $"\tSending {numberOfSignals} IR signals via the async output manager..." );
            for ( var i = 0; i < numberOfSignals; i++ )
            {
                // Queue the 50 signals for output. Just use one output port for each
                // signal, though we can send to more if wanted.
                var opList = new OutputList
                {
                    new OutputPort
                    {
                        PortNumber = rand.Next( 1, irNBIII.NumIROutputs + 1 ),
                        PowerLevel = 50, // Half power
                    }
                };
                // Queue each signal for output.
                asynOutputManager.QueueSignal( GetSignal(), opList );
            }

            // Wait by looping. A bit ugly but for the demo we want to know when the output of
            // this lot is finished.
            while ( asynOutputManager.QueueState.Length > 0 ) Thread.Sleep( 50 );
            // Wait a wee bit longer so that all output completes
            Thread.Sleep( 500 );

            irNBIV?.DisplayText( "" );

            Console.WriteLine( "\tDone." );
        }

        /// <summary>
        /// Uses the standard async API to send IR signals to different outputs.
        /// </summary>
        public void DoAsyncAPICalls( IRedRat4_III irNBIII )
        {
            const int numberOfSignals = 20;        // Output this many IR signals on each port.

            // Display info if MK-IV
            var irNBIV = irNBIII as IRedRat4_IV;
            if ( irNBIV == null )
            {
                Console.Error.WriteLine("ERROR: Device is not a MK-IV irNetBox.");
                return;
            }

            irNBIV.DisplayText("Async o/p V2");

            Console.WriteLine( "\nASYNC API CALLS" );
            Console.WriteLine( $"\tStarting {irNBIII.NumIROutputs} threads to output {numberOfSignals} IR signals on each port." );

            const int outputPowerLevel = 45;

            // Create our thread sync object.
            finished = new ManualResetEvent( false );
            Interlocked.Increment( ref threadCount );

            // What we do here is start a new thread for each port, in effect having 16 different
            // entities sending IR data at the same time.
            for ( var op = 0; op < irNBIII.NumIROutputs; op++ )
            {
                Interlocked.Increment( ref threadCount );       // Increment our thread counter.
                var opWorker = new IrNbIIISignalOutputWorker( irNBIII, op + 1, GetSignal, 
                    outputPowerLevel, numberOfSignals, WorkerThreadCompleted );
                ThreadPool.QueueUserWorkItem( opWorker.SignalOutputLoop );
            }

            // Indicate that we've completed.
            WorkerThreadCompleted();
            // Wait for the worker threads to finish...
            finished.WaitOne();

            irNBIV.DisplayText("");

            Console.WriteLine( "\tDone." );
        }

        /// <summary>
        /// Called by each worker thread (and main thread) once complete.
        /// </summary>
        public void WorkerThreadCompleted()
        {
            if ( Interlocked.Decrement( ref threadCount ) == 0 )
            {
                // All workers finished, so let the main thread continue.
                finished.Set();
            }
        }

        /// <summary>
        /// Returns a random IR signal for output.
        /// </summary>
        public ModulatedSignal GetSignal()
        {
            // Load device DB containing IR signals if have not already done so.
            if ( deviceDB == null )
            {
                FileStream fs = null;
                try
                {
                    var ser = new XmlSerializer( typeof( AVDeviceDB ) );
                    fs = new FileStream( new FileInfo( dbFilename ).FullName, FileMode.Open );
                    deviceDB = (AVDeviceDB)ser.Deserialize( fs );
                }
                finally
                {
                    fs?.Close();
                }                
            }

            // Get a random IR signal
            ModulatedSignal signal = null;
            // Not all IR signals read from the DB are modulated signals, so loop until we get one.
            while ( signal == null )
            {
                var index = rand.Next( deviceDB.AVDevices.Length );
                var device = deviceDB.AVDevices[index];
                signal = device.Signals[rand.Next( device.Signals.Length )] as ModulatedSignal;
            }
            return signal;
        }
    }
}
